/*
 * Copyright 2022-2023 Sensaru GmbH
 */

#ifndef LIBC1_MODULE_C1CORECLIENT_H
#define LIBC1_MODULE_C1CORECLIENT_H

#include "AssetId.h"
#include "Acl/UserAcl.h"

#include <homegear-base/BaseLib.h>
#include <c1-net/TcpSocket.h>
#include <c1-util/Queue.h>
#include <shared_mutex>

#include <gnutls/gnutls.h>
#include <gnutls/abstract.h>

namespace C1Module {

class CoreClientException : public BaseLib::Exception {
 public:
  explicit CoreClientException(const std::string &message) : Exception(message) {}
};

enum class PrincipalType {
  kUnknown = 0,
  kSuperuser = 1,
  kSystemProvider = 2,
  kSystemDistributor = 3,
  kBusinessPartner = 4,
  kEnduser = 5,
  kHomeClient = 6,
  kModule = 7,
  kEvent = 8
};

enum class LocationIdFilterTypes {
  kExact,
  kParent
};

struct LocationIdFilter {
  LocationIdFilterTypes filter_type = LocationIdFilterTypes::kExact;
  AssetId id;
};

struct ResultingPrincipal {
  PrincipalType type = PrincipalType::kUnknown;

  /**
   * The resulting ID of the principal. `sp`, `sd` and `bp` must always be set - even for system distributor, system provider or super users.
   */
  PrincipalId principal_id;

  /**
   * Either the user's or the home client's id.
   */
  std::string id;

  /**
   * Only set for users
   */
  std::string email;
};

struct CoreClientInfo {
  std::string node_name;
  std::string module_name;
  std::string module_description;
  std::string module_ui_menu;
  std::string module_ui_category;
  std::string module_app_ui_url;
  std::string module_app_ui_category;
  BaseLib::PVariable module_event_subscriptions = std::make_shared<BaseLib::Variable>(BaseLib::VariableType::tArray);
  BaseLib::PVariable acl_info = std::make_shared<BaseLib::Variable>(BaseLib::VariableType::tStruct);
  BaseLib::PVariable widget_info = std::make_shared<BaseLib::Variable>(BaseLib::VariableType::tStruct);
  bool allow_business_partner_user_access = true;
  bool allow_end_user_access = false;
  bool allow_home_client_access = false;
  /*
   * Allows access to system providers only.
   */
  bool system_provider_module = false;
  std::string ca_file;
  std::string client_cert_file;
  std::string client_key_file;
  std::function<void(uint32_t logLevel, const std::string &message)> log_callback;
  std::function<BaseLib::PVariable(const std::string &methodName,
                                   const BaseLib::PArray &parameters,
                                   const ResultingPrincipal &resulting_principal,
                                   const PUserAcl &acl,
                                   const BaseLib::PVariable &metadata)> rpc_method_callback;
  std::function<void(const std::string &hostname, uint16_t port)> connected_callback;
  std::function<void()> disconnected_callback;
  bool debug_output = false;
  uint32_t request_processing_thread_count = 50;
};

class CoreClient final : public C1Util::Queue {
 public:
  /**
   * Constructor
   *
   * @param base_lib The base library object
   * @param max_content_size The maximum size of RPC responses in bytes. Packets larger than this are dropped. There will be a warning when 80 % of this size is reached by a packet.
   */
  [[maybe_unused]] explicit CoreClient(size_t max_content_size = 0);
  ~CoreClient() override;

  std::pair<uint32_t, uint32_t> QueueSizes();

  /**
   * Returns `true` when Connected to c1-core.
   *
   * @return
   */
  bool Connected();

  /**
   * Starts the c1-core client and connects to an c1-core instance. You do not need to worry about disconnects and reconnecting to c1-core. This is all handled by this class.
   *
   * @param seed_servers
   * @param core_client_info
   */
  void Start(const std::unordered_map<std::string, uint16_t> &seed_servers, const CoreClientInfo &core_client_info);

  /**
   * Disconnects from c1-core and stops the c1-core client.
   */
  void Stop();

  /**
   * Calls an RPC method on c1-core.
   *
   * @param method_name
   * @param parameters
   * @param metadata
   * @param wait
   * @param timeout
   * @return The result of the RPC call.
   *
   * @throws CoreClientException Thrown on any error during the method call.
   */
  BaseLib::PVariable Invoke(const std::string &method_name, const BaseLib::PArray &parameters, BaseLib::PVariable &metadata, bool wait = true, int32_t timeout = 3600000);

  /**
   * Shortcut to call publish on a module.
   *
   * @param topic
   * @param value
   * @param qos 0 or 1
   * @return Returns the result of the call to invoke. When qos is set to 0, the method returns immediately.
   *
   * @throws CoreClientException Thrown on any error during the method call.
   */
  BaseLib::PVariable Publish(const std::string &topic, const BaseLib::PVariable &value, int32_t qos);

  /**
   * Shortcut to call moduleMethod on a module.
   *
   * @param module_name
   * @param method_name
   * @param parameters
   * @param token A unique token to match the response.
   * @param qos
   * @return Returns the token used for the call. This token can be used to match the response call to moduleMethodResponse.
   *
   * @throws CoreClientException Thrown on any error during the method call.
   */
  std::string ModuleMethod(const std::string &module_name,
                           const std::string &method_name,
                           const BaseLib::PVariable &parameters,
                           const std::string &token = "",
                           int32_t qos = 0,
                           const BaseLib::PVariable &user_id = std::make_shared<BaseLib::Variable>(BaseLib::VariableType::tStruct));
 private:
  struct AverageMeanData {
    AverageMeanData() = default;

    AverageMeanData(const AverageMeanData &data) {}

    std::atomic<int64_t> last_measurement{0};
    std::atomic<double> last_output{0.0};
  };

  struct RequestInfo {
    std::mutex wait_mutex;
    std::condition_variable condition_variable;
  };
  typedef std::shared_ptr<RequestInfo> PRequestInfo;

  enum class QueueEntryType {
    request,
    response
  };

  class QueueEntry : public C1Util::QueueEntry {
   public:
    QueueEntryType queue_entry_type = QueueEntryType::request;
  };

  class RequestQueueEntry : public QueueEntry {
   public:
    std::string method_name;
    BaseLib::PArray parameters;
  };

  class ResponseQueueEntry : public QueueEntry {
   public:
    BaseLib::PVariable response;
  };

  struct RpcResponse {
    std::atomic_bool finished{false};
    int32_t packet_id = 0;
    BaseLib::PVariable response;
    BaseLib::PVariable metadata;
  };

  typedef std::shared_ptr<RpcResponse> PRpcResponse;

  const int64_t start_time_ = BaseLib::HelperFunctions::getTime();
  std::atomic_bool started_ = false;
  std::string module_id_;
  std::string module_system_provider_id_;
  std::string module_system_distributor_id_;
  std::string module_business_partner_id_;
  CoreClientInfo core_client_info_;

  // {{{ Metrics
  AverageMeanData thread_load_average_;
  AverageMeanData packets_per_minute_received_;
  AverageMeanData packets_per_minute_sent_;
  AverageMeanData request_processing_time_;
  std::atomic<double> max_request_processing_time_current_{0.0};
  std::atomic<double> max_request_processing_time_{0.0};
  std::atomic<double> max_latency_current_{0.0};
  std::atomic<double> max_latency_{0.0};
  std::atomic<int64_t> last_reconnect_{0};

  std::atomic<int64_t> last_cycle_{0};
  // }}}

  // {{{ Listener thread
  std::atomic_bool stop_listen_thread_{false};
  std::thread listen_thread_;
  // }}}

  // {{{ Worker thread
  std::atomic_bool stop_worker_thread_{false};
  std::thread worker_thread_;
  // }}}

  // {{{ Socket
  /**
   * Not protected by mutex as it is only used in Connect() which in return is only called from Listen().
   */
  std::mutex seed_servers_mutex_;
  std::unordered_map<std::string, uint16_t> seed_servers_;
  std::unordered_map<std::string, uint16_t> seed_servers_backup_;
  std::shared_ptr<C1Net::TcpSocket> core_socket_;
  std::atomic_bool reconnect_socket_{false};
  // }}}

  // {{{ Needed for "invoke()"
  std::mutex request_info_mutex_;
  std::map<pthread_t, PRequestInfo> request_info_;
  std::atomic<int32_t> current_packet_id_{0};
  std::mutex rpc_responses_mutex_;
  std::unordered_map<pthread_t, std::unordered_map<int32_t, PRpcResponse>> rpc_responses_;
  // }}}

  // {{{ Data processing
  size_t max_content_size_ = 0;
  std::unique_ptr<BaseLib::Rpc::RpcEncoder> rpc_encoder_;
  std::unique_ptr<BaseLib::Rpc::RpcDecoder> rpc_decoder_;
  // }}}

  // {{{ RPC methods
  std::map<std::string, std::function<BaseLib::PVariable(const BaseLib::PArray &parameters, const PUserAcl &acl)>> local_rpc_methods_;
  // }}

  /**
   * Connects to c1-core. Must only be called from one thread (Listen()), because associated variables are not protected by a mutex.
   */
  void Connect();
  void Listen();

  /**
   * Reads a response from _cloudSocket. Make sure it's mutex is locked before calling this method.
   *
   * @return
   */
  BaseLib::PVariable ReadRpcResponse(pthread_t thread_id, int32_t packet_id);
  void ProcessQueueEntry(int32_t index, std::shared_ptr<C1Util::QueueEntry> &entry) override;
  BaseLib::PVariable ProcessRequest(const std::string &method_name, BaseLib::PArray &parameters);
  void ProcessResponse(const BaseLib::PVariable &response);
  void Worker();
  void UpdateAverages();
  void PublishStatus();

  void Send(const std::vector<uint8_t> &data);
  void Send(const uint8_t *buffer, std::size_t bytes_to_write);

  // {{{ RPC methods
  BaseLib::PVariable Ping(const BaseLib::PArray &parameters);
  // }}}
};

}

#endif //LIBC1_MODULE_C1CORECLIENT_H
